# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to check model compatibility for CNN2SNN conversion.
"""
from keras import layers
from quantizeml import layers as qlayers

from .blocks import split_model_into_blocks
from ..transforms.sequential import _check_layers_data_format, _check_layer_inbounds
from ..compatibility_checks import _check_reshape_layer
from ..akida_versions import get_akida_version, AkidaVersion

neural_layers = (qlayers.QuantizedConv2D, qlayers.QuantizedSeparableConv2D,
                 qlayers.QuantizedDense, qlayers.QuantizedDepthwiseConv2D,
                 qlayers.QuantizedConv2DTranspose, qlayers.QuantizedDepthwiseConv2DTranspose)
skippable_layers = (layers.InputLayer, layers.Reshape, layers.Flatten, layers.Rescaling,
                    layers.Activation, layers.Softmax, layers.Dropout)
pooling_layers = (qlayers.QuantizedGlobalAveragePooling2D, qlayers.QuantizedMaxPool2D)
norm_layers = (qlayers.LayerMadNormalization, qlayers.QuantizedBatchNormalization)
activation_layers = (qlayers.QuantizedReLU,)
stem_layers = (qlayers.ClassToken, qlayers.AddPositionEmbs)


class CompatibilityMessage:
    """Helper to represent an error message

    Args:
        message (str): main error message.
        layers (list or tf.keras.Layers): layer(s) that will raise(s) the exception.
            Defaults to None.
    """

    def __init__(self, message, layers=None):
        if not message.endswith("."):
            message += "."
        self.message = message
        # Replace empty list by None
        self.layers = layers or None

    def __str__(self):
        if self.layers is None:
            return self.message

        message = f"{self.message} Layer(s): "
        if isinstance(self.layers, list):
            message += str([x.name for x in self.layers])
        else:
            message += self.layers.name
        return message


class CompatibilityBlockError:
    """Helper to raise a list of ``CompatibilityMessage`` errors"""

    def __init__(self):
        self._errors_msg = []

    @property
    def status(self):
        return len(self._errors_msg) > 0

    def push(self, message, layers=None):
        self._errors_msg.append(CompatibilityMessage(message, layers))

    def __str__(self):
        err_msg = "The following errors were found when the block was checked:"
        for err in self._errors_msg:
            err_msg += "\n\t- " + str(err)
        return err_msg

    def __call__(self):
        if self.status:
            raise RuntimeError(str(self))


def check_model_compatibility(model):
    r"""Checks if a QuantizeML model is compatible for Akida conversion.

    This function does NOT:

    - convert the QuantizeML model to an Akida model,
    - check if the model is compatible with Akida hardware

    It ONLY checks if the model design is compatible with Akida.

    Args:
        model (:obj:`tf.keras.Model`): the model to check.

    Returns:
        list: a list of sequences of layers ('blocks').
    """
    # Check general rules about model in three steps:
    # 1. Check if model has only one input and one output,
    # 2. Check right data format and
    # 3. Over Akida 1.0, check if model is sequential.
    _check_model_input_output(model)
    _check_layers_data_format(model)
    if get_akida_version() == AkidaVersion.v1:
        _check_layer_inbounds(model)

    # Split model into theirs blocks:
    blocks = split_model_into_blocks(model)

    # Evaluate block-by-block integrity
    for block in blocks:
        # Check reshape/flatten layers
        skippable, block = _extract_skippable_layers(block)
        _check_skippable_compatibility(skippable)
        # Check pooling layers
        _check_pooling_compatibility(block)
        # Check other rules
        _check_block_integrity(block)
    return blocks


def _split_block_into_layer_types(block):
    """Split input block into the following categories:

    - neural (processing) layers,
    - activation layers,
    - normalization layers and
    - stem block layers.

    Args:
        block (list): block layers to split.

    Returns:
        list: block divided into categories.
    """
    neurals, acts, norms, stems = [], [], [], []
    for layer in block:
        if isinstance(layer, neural_layers):
            neurals.append(layer)
        elif isinstance(layer, activation_layers):
            acts.append(layer)
        elif isinstance(layer, norm_layers):
            norms.append(layer)
        elif isinstance(layer, stem_layers):
            stems.append(layer)
        elif not isinstance(layer, pooling_layers):
            raise RuntimeError(f"Unrecognizable layer in block: {layer.name}")
    return neurals, acts, norms, stems


def _check_block_integrity(block):
    """Check integrity of one block.

    Args:
        block (list(tf.keras.Layer)): block to analysis.
    """
    if len(block) == 0 or isinstance(block[0], qlayers.Dequantizer):
        return

    def _islast(layer):
        outbound = layer.outbound_nodes
        if len(outbound) == 1:
            outbound_layer = outbound[0].outbound_layer
            if isinstance(outbound_layer, skippable_layers):
                return _islast(outbound_layer)
            # If outbound layer is a Dequantizer, target is the final quantized layer
            if isinstance(outbound_layer, qlayers.Dequantizer):
                return True
        return len(outbound) == 0

    def _isfirst(layer):
        ly_inbounds = layer.inbound_nodes[0].inbound_layers
        if not isinstance(ly_inbounds, (tuple, list)):
            ly_inbounds = [ly_inbounds]
        if len(ly_inbounds) == 1 and isinstance(ly_inbounds[0], skippable_layers):
            return _isfirst(ly_inbounds[0])
        return len(ly_inbounds) == 0

    # Start with Akida 1.0 restrictions
    errors = CompatibilityBlockError()
    ak_version = get_akida_version()
    if ak_version == AkidaVersion.v1:
        islast = _islast(block[-1])
        if len(block) == 1 and not islast:
            errors.push("In Akida 1.0, every block must have at least "
                        "one neural and activation layer")
        if islast and not isinstance(block[0], neural_layers):
            errors.push("The last block must start by a neural layer", block[0])
        if not (islast or isinstance(block[-1], activation_layers)):
            errors.push("A block must end by an activation layer", block[-1])

    # General limitations
    if len(block) > 1:
        neurals, acts, norms, stems = _split_block_into_layer_types(block)
        if len(neurals) != 1:
            errors.push("A block must have one neural layer", neurals)
        if block[0] != neurals[0]:
            errors.push("A block must start by a neural layer", block[0])
        if len(acts) > 1:
            errors.push("A block could have up to only one activation layer", acts)
        if len(norms) != 0:
            errors.push("A block should not have normalization layers: "
                        "these must be folded into previous neural layers", norms)
        if len(stems) > 0:
            if ak_version == AkidaVersion.v1:
                errors.push("Stem-block is not supported on Akida 1.0.")
            if not _isfirst(block[0]):
                errors.push("Stem-block must be the first one in the model.")
            if not all(isinstance(x, qlayers.ClassToken) for x in stems[1:-1]):
                errors.push("'ClassToken' is the only intermediate layer allowed "
                            "in a stem block.", stems[1:-1])
            if not isinstance(block[0], qlayers.QuantizedConv2D):
                errors.push("Stem-block must begin by one 'Conv2D'", block[0])
            if not isinstance(stems[-1], qlayers.AddPositionEmbs):
                errors.push("Stem-block must end by one 'AddPositionEmbs'.", stems[-1])

    # Raise exception if there is at least one error message
    errors()


def _check_model_input_output(model):
    """Asserts that model inputs/outputs are supported for conversion.

    The Keras model must have only one input layer and one output layer.
    On Akida 1.0, the input shape must 4-D (N, H, W, C).

    Args:
        model (tf.keras.model): the Keras model to check.
    """

    # Error if multiple inputs
    if len(model.input_names) > 1:
        raise RuntimeError("Model must have only one input layer. Receives "
                           f"inputs {model.input_names}.")

    # Error if multiple outputs
    if len(model.output_names) != 1:
        raise RuntimeError("Model must have only one output layer. Receives"
                           f"outputs {model.output_names}.")

    # Error if input shape is not 2D or 4D
    if len(model.input_shape) not in (2, 4) and get_akida_version() == AkidaVersion.v1:
        raise RuntimeError(
            "Input shape of model must be 2-D or 4-D (batch size + 1-D or 3-D "
            f"tensors). Receives input shape {model.input_shape}.")


def _extract_skippable_layers(block):
    """Split block into skippable and non skippable layers

    Args:
        block (tf.keras.Layer): block to split.

    Returns:
        tuple: list of skippable and non skippable layers
    """
    skippable, non_skippable = [], []
    for layer in block:
        if isinstance(layer, skippable_layers):
            skippable.append(layer)
        else:
            non_skippable.append(layer)
    return skippable, non_skippable


def _check_skippable_compatibility(skippables):
    """Check rules over quantized skippable layers

    Args:
        skippables (list(tf.keras.Layer)): layers list to check.
    """
    def _find_next_non_skippable_outbounds(layer):
        outbound = layer.outbound_nodes
        if len(outbound) == 1:
            outbound_layer = outbound[0].outbound_layer
            if isinstance(outbound_layer, skippable_layers):
                return _find_next_non_skippable_outbounds(outbound_layer)
            # If outbound layer is a Dequantizer, target is considered as the final quantized layer
            if isinstance(outbound_layer, qlayers.Dequantizer):
                outbound = []
        return outbound

    for layer in skippables:
        if isinstance(layer, qlayers.QuantizedReshape):
            # There are not rules on Akida 2.0 about QuantizedReshape.
            if get_akida_version() == AkidaVersion.v1:
                _check_reshape_layer(layer)
            continue
        # For the next layer types, we can skip checks if they are the last ones in the model
        outbound = _find_next_non_skippable_outbounds(layer)
        if len(outbound) == 0:
            continue
        elif isinstance(layer, qlayers.QuantizedFlatten):
            # QuantizedFlatten outbound must be one QuantizedDense
            outbound_layer = outbound[0].outbound_layer
            if len(outbound) > 1 or not isinstance(outbound_layer, qlayers.QuantizedDense):
                raise RuntimeError("A QuantizedFlatten layer is only supported at the end of "
                                   "the model or before a QuantizedDense. "
                                   f"Layer: {layer.name}")
        elif not isinstance(layer, skippable_layers):
            raise RuntimeError(f"Incompatible {layer.name} skippable layer.")


def _check_pooling_compatibility(block):
    """Asserts pooling layer is compatible for conversion.

    Two checks are performed here:
    - a pooling layer must have a Conv2D type as processing layer and
    - the padding of MaxPool2D must be the same as the padding of neural layer

    Args:
        block (list): the layers block.
    """
    errors = CompatibilityBlockError()
    conv_neural_layers = (qlayers.QuantizedConv2D,
                          qlayers.QuantizedSeparableConv2D, qlayers.QuantizedDepthwiseConv2D)
    pool_layer = [ly for ly in block if isinstance(ly, pooling_layers)]
    if len(pool_layer) == 0:
        return
    elif len(pool_layer) != 1:
        errors.push("A block could have up to only one pooling layer", pool_layer)

    # Raise error if pooling does not have a convolutional layer in the block
    pool_layer = pool_layer[0]
    conv_layer = [ly for ly in block if isinstance(ly, conv_neural_layers)]
    if len(conv_layer) != 1:
        errors.push("Pooling layer must be placed after a convolutional layer", conv_layer)

    if isinstance(pool_layer, qlayers.QuantizedGlobalAveragePooling2D) and pool_layer != block[-1] \
            and get_akida_version() == AkidaVersion.v2:
        errors.push("In v2, GAP should occur after ReLU and placed at the end of the block",
                    pool_layer)
    # Raise the exceptions(if exist)
    errors()

    # Raises error if the padding of MaxPool2D is different from the padding
    # of the neural processing layer.
    conv_layer = conv_layer[0]
    neur_pad = conv_layer.padding
    pool_pad = getattr(pool_layer, "padding", "")
    if isinstance(pool_layer, qlayers.QuantizedMaxPool2D) and neur_pad != pool_pad:
        raise RuntimeError(f"Pooling layer {pool_layer.name} (padding: {pool_pad}) must have "
                           f"the same padding as {conv_layer.name} (padding: {neur_pad}).")



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert Dequantizer to Akida.
"""
import numpy as np

import akida
from quantizeml.layers import Dequantizer

from .weights import broadcast_and_set_variable
from ..akida_versions import AkidaVersion, get_akida_version


def _set_dequantizer_variables(layer_ak, dequantizer):
    """Computes and sets the dequantizer recorded variables into the akida layer

    Args:
        layer_ak (:obj:`akida.Layer`): the targeted akida layer.
        dequantizer (:obj:`quantizeml.Dequantizer`): the source quantizeml layer.
    """
    assert isinstance(dequantizer, Dequantizer)

    # Extract the Dequantizer variables
    frac_bits = dequantizer.frac_bits
    scales = dequantizer.scales
    if isinstance(frac_bits, (list, tuple)):
        if len(frac_bits) > 1:
            raise RuntimeError(f"Multi-inbounds in {dequantizer.name} is not supported.")
        frac_bits = frac_bits[0]
        scales = scales[0] if scales else None

    # We project the frac_bits into the scales as:
    #   new_scales = scales / 2 ** frac_bits
    scales = scales.value if scales else 1.0
    scales /= 2**frac_bits.value
    scales = scales.numpy().astype(np.float32)

    # Depending on ak_version:
    layer_variables = layer_ak.variables
    if get_akida_version() == AkidaVersion.v1:
        # We should not modify the activation step if the model ends with an activation
        if layer_ak.parameters.activation == 1:
            return
        # Project scales in the final layer activation step
        params = ("act_step", layer_variables["act_step"] / scales)
    else:
        # Set scales in the independant layer
        assert layer_ak.parameters.layer_type == akida.LayerType.Dequantizer
        params = ("scales", scales)
    broadcast_and_set_variable(layer_variables, *params)


def convert_dequantizer(model_ak, layer_k):
    """Converts Dequantizer layer and set its variables into the Akida model.

    Args:
        model_ak (:obj:`akida.Model`): the model where the layer will be added.
        layer_k (:obj:`tf.keras.Layer`): the quantizeml layer to convert.
    """
    if not isinstance(layer_k, Dequantizer):
        raise TypeError(f"Layer {layer_k.name} was expected to be Dequantizer")

    # Depending on ak_version:
    if get_akida_version() == AkidaVersion.v1:
        # Take the last layer to set variables
        layer_ak = model_ak.layers[-1]
    else:
        # Create and add one akida.Dequantizer to the model
        layer_ak = akida.Dequantizer(name=layer_k.name)
        model_ak.add(layer_ak)

    # Set variables into the akida layer
    _set_dequantizer_variables(layer_ak, layer_k)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert QuantizedDense to Akida.
"""
from akida import LayerType, Dense1D, Dense2D, FullyConnected
from quantizeml.layers import (QuantizedDense, QuantizedReLU, QuantizedReshape, QuantizedFlatten,
                               WeightQuantizer, AlignedWeightQuantizer)
import numpy as np

from .layer_utils import get_inbound_layers
from .weights import broadcast_and_set_variable
from ..akida_versions import AkidaVersion, get_akida_version
from .activations import parse_relu, set_relu_variables
from .outputs import set_output_variables


def _set_dense2d_variables(ak_layer, dense, block_input_shape):
    """Computes and sets the variables for an Akida Dense1D/Dense2D layer.

    This function converts the variables of a block of Keras layers and sets them into
    the corresponding variables of the equivalent Akida layer.

    Args:
        ak_layer (:obj:`akida.Layer`): the targeted akida layer.
        dense (:obj:`tf.keras.Layer`): the keras dense layer.
        block_input_shape (list): the dense block input shape.
    """
    type_ak = ak_layer.parameters.layer_type
    assert isinstance(dense, QuantizedDense)
    assert type_ak in [LayerType.Dense2D, LayerType.FullyConnected]
    assert isinstance(dense.weight_quantizer, WeightQuantizer)

    variables_ak = ak_layer.variables

    # get the QuantizedDense weights
    weights_ak = dense.weight_quantizer.qweights.value.fp.values.numpy()
    # get the QuantizedDense bias and shift
    if dense.use_bias:
        bias_quantizer = dense.bias_quantizer
        assert isinstance(bias_quantizer, AlignedWeightQuantizer)
        bias = bias_quantizer.qweights.value.values.numpy().astype(np.int32)
        bias_shift = bias_quantizer.shift.value.numpy().astype(np.uint8)
        if type_ak == LayerType.Dense2D:
            variables_ak["bias"] = (bias >> bias_shift).astype(np.int8)
            broadcast_and_set_variable(variables_ak, "bias_shift", bias_shift)
        else:
            # Store bias into the threshold variable
            variables_ak["threshold"] = -bias

    input_shift = getattr(dense, 'input_shift', None)
    if input_shift is not None and type_ak != LayerType.FullyConnected:
        broadcast_and_set_variable(variables_ak, "input_shift",
                                   input_shift.value.numpy().astype(np.uint8))

    if type_ak == LayerType.FullyConnected:
        # In AkidaVersion.v1, it is only possible to convert flat inputs
        input_shape = dense.input_shape
        units = ak_layer.parameters.units
        if len(input_shape) != 2:
            raise RuntimeError(f"Expected a flat input, received {input_shape}. Try to include a "
                               f"Reshape or Flatten layer before of {dense.name}.")
        if block_input_shape != input_shape:
            _, X, Y, C = block_input_shape
            # When a fully connected layer follows a convolutional layer, we need to modify
            # the way the weights are laid out because the Keras and Akida flatten operations
            # that happen on spatial dimensions are inverted (row-major versus col-major).
            # We therefore need to:
            # - reshape the Keras (FxN) weights to match the block input shape (XxYxCxN),
            # - transpose to the equivalent akida shape (CxYxXxN).
            weights_ak = weights_ak.reshape(X, Y, C, units).transpose(2, 1, 0, 3)
        weights_ak = weights_ak.reshape(1, 1, input_shape[-1], units)
    variables_ak["weights"] = weights_ak.astype(np.int8)


def _parse_dense2d(layer):
    """Parses a quantizeml.QuantizedDense parameters.

    Args:
        layer (:obj:`tf.keras.Layer`): the layer to parse.

    Returns:
        dict: the corresponding akida parameters.
    """

    assert isinstance(layer, QuantizedDense)
    weight_bits = layer.weight_quantizer.bitwidth
    # The only weight bitwidth supported is 4 or 8
    assert weight_bits in [4, 8]

    # In quantizeml one bit is reserved for the sign in the buffer bitwidth
    # variable, but in akida this value has to be added back to have the
    # correct clipping.
    buffer_bits = layer.buffer_bitwidth + 1

    # Find out if there is a quantizer
    out_quantizer = getattr(layer, "out_quantizer", None)

    dense_params = dict(units=layer.units, activation=False, name=layer.name)

    if get_akida_version() == AkidaVersion.v1:
        assert weight_bits == 4
        dense_params["weights_bits"] = weight_bits
    else:
        dense_params["buffer_bits"] = buffer_bits
        dense_params["output_bits"] = out_quantizer.bitwidth if out_quantizer else buffer_bits

    return dense_params


def convert_dense_block(model_ak, layers):
    """Converts a dense block into an akida dense layer.

    The expected sequence is:

        - QuantizedDense,
        - QuantizedReLU (optional).

    Args:
        model_ak (:obj:`akida.Model`): the Akida model where the model will be added.
        layers (list(:obj:`tf.keras.Layer`)): the remaining model layers to convert.

    Return:
        int: the number of layers in the block or O if the first layer is not a QuantizedDense.
    """

    if len(layers) == 0:
        return 0

    # Initialize an empty list
    dense_block = []

    index = 0
    block_input_shape = layers[0].input_shape
    while isinstance(layers[index], (QuantizedReshape, QuantizedFlatten)):
        dense_block.append(layers[index])
        index += 1

    # at this point the layer dense should be:
    dense = layers[index]
    if not isinstance(dense, QuantizedDense):
        return 0

    # add the main dense layer to it
    dense_block.append(dense)

    # by default the dense block should be converted in an Akida Dense2D layer. But if the dense
    # input has a flattened shape or was flattened before with a Reshape or a Flatten layer then
    # it's converted to an Akida Dense1D. To check if the input was flattened, one assert that
    # the dense layer input_shape is 2 (batch_size, flattened_dim).
    is_dense1d = False
    if len(dense.input_shape) == 2:
        is_dense1d = True

    dense_params = _parse_dense2d(dense)

    # Find out if there is a quantizer
    out_quantizer = getattr(dense, "out_quantizer", False)

    relu_layer = None
    relu_index = len(dense_block)
    if len(layers) > relu_index and isinstance(layers[relu_index], QuantizedReLU):
        if out_quantizer:
            raise RuntimeError(f"{dense.name} layer is followed by an activation layer.\
                It should not have an output_quantizer.")
        relu_layer = layers[relu_index]
        dense_block.append(relu_layer)
        act_params = parse_relu(relu_layer)
        # Replace previous output_bits with relu output quantizer value.
        dense_params.update(act_params)

    # Create the Akida Dense2D or Dense1D layer, following the AkidaVersion
    if get_akida_version() == AkidaVersion.v2:
        dense_ak = Dense1D(**dense_params) if is_dense1d else Dense2D(**dense_params)
        # Retrieve the akida inbound layers
        inbound_layers_ak = get_inbound_layers(model_ak, dense)
        # Add it to the model
        model_ak.add(dense_ak, inbound_layers_ak)
    else:
        dense_ak = FullyConnected(**dense_params)
        model_ak.add(dense_ak)

    # Set the main Dense2D variables
    _set_dense2d_variables(dense_ak, dense, block_input_shape)

    # Set the optional activation variables
    if relu_layer:
        set_relu_variables(dense_ak, relu_layer)
        # the effective output_quantizer should be the relu one
        out_quantizer = getattr(relu_layer, "out_quantizer", False)

    # Set the optional output_quantizer variables
    if out_quantizer:
        set_output_variables(dense_ak, out_quantizer)

    return len(dense_block)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert QuantizedBatchNormalization to Akida.
"""
from akida import LayerType, BatchNormalization
import quantizeml.layers as qlayers
import numpy as np

from .weights import broadcast_and_set_variable
from .outputs import set_output_variables
from .activations import parse_relu, set_relu_variables
from .layer_utils import get_inbound_layers


def _set_batchnorm_variables(ak_layer, block):
    """Computes and sets the variables for an Akida BatchNormalization layer.

    This function converts the variables of a Keras layer and sets them into
    the corresponding variables of the equivalent Akida layer.

    Args:
        ak_layer (:obj:`akida.Layer`): the targeted akida layer.
        block (list(:obj:`tf.keras.Layer`)): the block of keras layers.
    """
    layer_bn = block[0]
    assert isinstance(layer_bn, qlayers.QuantizedBatchNormalization)
    assert ak_layer.parameters.layer_type == LayerType.BatchNormalization
    assert isinstance(layer_bn.a_quantizer, qlayers.WeightQuantizer)
    assert isinstance(layer_bn.b_quantizer, qlayers.AlignedWeightQuantizer)

    # get the QuantizedBatchNormalization a, b and shift
    a_ak = layer_bn.a_quantizer.qweights.value.values.numpy()
    b_quantizer = layer_bn.b_quantizer
    b = b_quantizer.qweights.value.values.numpy().astype(np.int32)
    b_shift = b_quantizer.shift.value.numpy().astype(np.uint8)
    b_ak = (b >> b_shift).astype(np.int8)

    variables_ak = ak_layer.variables

    input_shift = layer_bn.input_shift.value
    if input_shift is not None:
        broadcast_and_set_variable(variables_ak, "input_shift",
                                   input_shift.numpy().astype(np.uint8))

    variables_ak["a"] = a_ak.astype(np.int8)
    variables_ak["b"] = b_ak
    broadcast_and_set_variable(variables_ak, "b_shift", b_shift)

    out_quantizer = getattr(layer_bn, "out_quantizer", False)
    if len(block) > 1:
        relu_layer = block[1]
        set_relu_variables(ak_layer, relu_layer)
        # the effective output_quantizer should be the relu one
        out_quantizer = getattr(relu_layer, "out_quantizer", False)

    if out_quantizer:
        set_output_variables(ak_layer, out_quantizer)


def _create_batchnorm(layer_block):
    """Parses a quantizeml QuantizedBatchNormalization layer and returns the
    params to create the corresponding Akida BatchNormalization layer.

    Args:
        layer (:obj:`tf.keras.Layer`): the quantizeml
            QuantizedBatchNormalization layer to convert.

    Returns:
        :obj:`akida.BatchNormalization`: The created akida layer.
    """
    layer = layer_block[0]
    assert isinstance(layer, qlayers.QuantizedBatchNormalization)
    act_params = {}
    if len(layer_block) > 1:
        act_params = parse_relu(layer_block[1])
    # In quantizeml one bit is reserved for the sign in the buffer bitwidth
    # variable, but in akida this value has to be added back to have the
    # correct clipping.
    buffer_bits = layer.buffer_bitwidth + 1
    # If there is no activation, find out output_bits
    if not act_params:
        out_quantizer = getattr(layer, "out_quantizer", False)
        if out_quantizer:
            act_params['output_bits'] = out_quantizer.bitwidth
        else:
            # Default to buffer bitwidth
            act_params['output_bits'] = buffer_bits

    return BatchNormalization(**act_params,
                              buffer_bits=buffer_bits,
                              name=layer.name)


def convert_batchnorm_block(model_ak, layers):
    """Converts QuantizedBatchNormalization layer and its variables and adds
    it to the Akida's model. If followed by QuantizedReLU, it will set its
    variables too.

    Args:
        model_ak (:obj:`akida.Model`): the Akida model where the model will be
            added.
        layers (list(:obj:`tf.keras.Layer`)): the remaining model layers to
            convert.

    Return:
        int: the number of converted layers.
    """
    if not layers or not isinstance(layers[0],
                                    qlayers.QuantizedBatchNormalization):
        return 0

    block = [layers[0]]
    if len(layers) > 1 and isinstance(layers[1], qlayers.QuantizedReLU):
        block.append(layers[1])
    layer_ak = _create_batchnorm(block)

    # Retrieve the akida inbound layers
    inbound_layers_ak = get_inbound_layers(model_ak, layers[0])
    model_ak.add(layer_ak, inbound_layers_ak)
    _set_batchnorm_variables(layer_ak, block)
    return len(block)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert Conv2DTranspose to Akida.
"""
import numpy as np

from quantizeml.layers import (QuantizedConv2DTranspose, QuantizedReLU,
                               QuantizedMaxPool2D, WeightQuantizer, AlignedWeightQuantizer)
from akida import Conv2DTranspose

from ..akida_versions import AkidaVersion, get_akida_version
from .activations import set_relu_variables
from .weights import broadcast_and_set_variable
from .outputs import set_output_variables
from .blocks import parse_block_additional_layers, get_block_out_quantizer


def _set_conv2d_transpose_variables(layer_ak, layer_k):
    """Computes and sets the variables for an Akida Conv2DTranspose layer.

    This function converts the variables of a Keras layer and sets them into
    the corresponding variables of the equivalent Akida layer.

    Args:
        layer_ak (:obj:`akida.Layer`): the targeted akida layer.
        layer_k (:obj:`tf.keras.Layer`): the source quantized layer.
    """
    assert isinstance(layer_k, QuantizedConv2DTranspose)
    assert isinstance(layer_k.weight_quantizer, WeightQuantizer)
    if layer_k.use_bias:
        assert isinstance(layer_k.bias_quantizer, AlignedWeightQuantizer)

    # Get the weights (Note that in qweights the filter and channel dimensions are already
    # transposed)
    weights = layer_k.weight_quantizer.qweights.value.fp.values.numpy()

    layer_ak.variables["weights"] = weights.astype(np.int8)

    # Set the bias (if there is one)
    if layer_k.use_bias:
        bias_quantizer = layer_k.bias_quantizer
        bias = bias_quantizer.qweights.value.values.numpy().astype(np.int32)
        bias_shift = bias_quantizer.shift.value.numpy().astype(np.uint8)

        # Unshift the bias and store it
        layer_ak.variables["bias"] = (bias >> bias_shift).astype(np.int8)
        # Also store the bias shift
        broadcast_and_set_variable(layer_ak.variables, "bias_shift", bias_shift)

    # Set input shift if available
    if getattr(layer_k, 'input_shift', None):
        broadcast_and_set_variable(layer_ak.variables, "input_shift",
                                   layer_k.input_shift.value.numpy().astype(np.uint8))


def _parse_conv2d_transpose(layer):
    """Parses a quantizeml.QuantizedConv2DTranspose parameters.

    Args:
        layer (:obj:`tf.keras.Layer`): the layer to parse.

    Returns:
        dict: the corresponding akida parameters.
    """
    assert isinstance(layer, QuantizedConv2DTranspose)

    # Make sure the QuantizedConv2DTranspose kernel size and stride params are square
    assert layer.kernel_size[0] == layer.kernel_size[1], (
        "QuantizedConv2DTranspose kernel_size should be square")
    assert layer.kernel_size[0] in (3, 4)
    assert layer.strides[0] == layer.strides[1], (
        "QuantizedConv2DTranspose strides should be the same on both dimensions.")

    assert layer.strides[0] == 2, ("QuantizedConv2DTranspose handles only stride 2")
    # Padding value must be built in constructor
    assert layer.padding == "same", ("QuantizedConv2DTranspose handles only 'same' padding")
    # The only weight bitwidth supported is [4, 8]
    weight_bits = layer.weight_quantizer.bitwidth
    assert weight_bits in [4, 8]

    # In quantizeml one bit is reserved for the sign in the buffer bitwidth
    # variable, but in akida this value has to be added back to have the
    # correct clipping.
    buffer_bits = layer.buffer_bitwidth + 1

    return dict(
        filters=layer.filters,
        kernel_size=layer.kernel_size[0],
        buffer_bits=buffer_bits,
        activation=False,
        name=layer.name
    )


def convert_conv_transpose_block(model_ak, layers):
    """Converts a convolutional transpose block into an akida Conv2DTranspose layer.

    The expected sequence is:

    - QuantizedConv2DTranspose,
    - QuantizedReLU (optional).

    Args:
        model_ak (:obj:`akida.Model`): the target Akida model.
        layers (list(:obj:`tf.keras.Layer`)): the remaining unconverted keras model layers.

    Returns:
        int: the number of layers in the block or 0 if the first layer is not a
        QuantizedConv2DTranspose.
    """
    if len(layers) == 0 or not isinstance(layers[0], QuantizedConv2DTranspose):
        return 0

    if get_akida_version() == AkidaVersion.v1:
        return 0

    conv_transpose = layers[0]

    # Evaluate the convolutional transpose layer parameters
    conv_transpose_params = _parse_conv2d_transpose(conv_transpose)

    # Identify the next layers
    next_layers = parse_block_additional_layers(layers, conv_transpose_params)

    if len(next_layers) > 1 or next_layers.get(QuantizedMaxPool2D):
        raise RuntimeError("QuantizedConv2DTranspose may only be followed by an optional"
                           " QuantizedReLU")

    # Check if we have ReLU
    relu_layer = next_layers.get(QuantizedReLU, None)

    # Get the layer block output_quantizer
    out_quantizer = get_block_out_quantizer([conv_transpose] + [*next_layers.values()])

    # add output quantizer bitwidth parameter
    if out_quantizer:
        conv_transpose_params["output_bits"] = out_quantizer.bitwidth
    else:
        conv_transpose_params["output_bits"] = conv_transpose_params["buffer_bits"]

    # Create Akida layer
    conv_transpose_ak = Conv2DTranspose(**conv_transpose_params)
    # Add layer to the model to build its internal variables
    model_ak.add(conv_transpose_ak)

    # Set base variables
    _set_conv2d_transpose_variables(conv_transpose_ak, conv_transpose)

    # Set optional activation variables
    if relu_layer:
        set_relu_variables(conv_transpose_ak, relu_layer)

    # Set optional output_quantizer variables
    if out_quantizer:
        set_output_variables(conv_transpose_ak, out_quantizer)

    return 1 + len(next_layers)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Utility function to convert block layers to akida.
"""
from quantizeml.layers import (QuantizedMaxPool2D,
                               QuantizedGlobalAveragePooling2D,
                               QuantizedReLU,
                               QuantizedExtractToken,
                               Dequantizer)

from .pooling import parse_max_pooling, parse_global_average_pooling
from .activations import parse_relu
from ..akida_versions import AkidaVersion, get_akida_version


def parse_block_additional_layers(layers, block_params):
    """Parse block additional layers into the parameters of one.

    In Akida v1, we are able to manage this sequence of layers:

    - QuantizedMaxPooling2D or QuantizedGlobalAveragePooling2D
    - QuantizedReLU.

    In v2, the possible sequences are:

    - QuantizedMaxPooling2D
    - QuantizedReLU

    or:

    - QuantizedReLU
    - QuantizedGlobalAveragePooling2D

    Args:
        layers (list(:obj:`tf.keras.Layer`)): the layers block.
        block_params (dict): the current block parameters.

    Returns:
        dict: additional layers found in the process, with layer type as key
    """
    # Identify the next layers
    next_layers = {}
    if get_akida_version() == AkidaVersion.v1:
        layer_types = [QuantizedMaxPool2D, QuantizedGlobalAveragePooling2D, QuantizedReLU]
    else:
        layer_types = [QuantizedMaxPool2D, QuantizedReLU, QuantizedGlobalAveragePooling2D]
    index = 1
    while index < len(layers) and layer_types:
        layer_type = layer_types.pop(0)
        layer = layers[index]
        if isinstance(layer, layer_type):
            next_layers[layer_type] = layer
            index += 1

    relu_layer = next_layers.get(QuantizedReLU, None)
    max_pool_layer = next_layers.get(QuantizedMaxPool2D, None)
    global_average_pool_layer = next_layers.get(QuantizedGlobalAveragePooling2D, None)

    if len(next_layers) == 1 and max_pool_layer:
        raise RuntimeError("Unsupported configuration: a maxpool requires an activation.")

    # Evaluate the neural layer parameters
    neural_name = layers[0].name
    if getattr(layers[0], "out_quantizer", False) and len(next_layers) > 0:
        raise RuntimeError(f"{neural_name} layer is followed by a pool/activation layer."
                           "It should not have an output_quantizer.")

    if max_pool_layer:
        pool_params = parse_max_pooling(max_pool_layer)
        # Padding checked on check_compatibility_model()
        block_params.update(pool_params)
    if global_average_pool_layer:
        pool_params = parse_global_average_pooling(global_average_pool_layer)
        block_params.update(pool_params)
    if relu_layer:
        act_params = parse_relu(relu_layer)
        block_params.update(act_params)

    return next_layers


def split_model_into_blocks(model):
    """Search into the model the sets of possible blocks and return them.

    A set of layers are considered as block if:
        1. There is only one output quantizer (in the last layer of the block)
        2. There are no multi-inbound layers
        3. There are no QuantizedExtractToken layers

    Layers with more than one inbound layer will be handled separately.

    Args:
        model (:obj:`tf.models.Model`): The model to split

    Returns:
        list: a list of sequences of layers ('blocks').
    """
    def _get_inbound_layers(target_layer):
        inbound = target_layer.inbound_nodes[0].inbound_layers
        return inbound if isinstance(inbound, (list, tuple)) else [inbound]

    def _check_single_layer_block(target_layer):
        # Single blocks will be multi-inputs and QuantizedExtractToken
        prev_layers = _get_inbound_layers(target_layer)
        if len(prev_layers) > 1:
            return True
        if isinstance(target_layer, QuantizedExtractToken):
            return True
        return False

    def _search_block(target_layer):
        # We consider the target is part of one block if there are previous
        # layers (target is not the model's input), or it's multi-inbound.
        prev_layers = _get_inbound_layers(target_layer)
        if len(prev_layers) != 1:
            return [target_layer]
        # If previous layer has an output quantizer, the blocks end here.
        next_target = prev_layers[0]
        out_quantizer = getattr(next_target, "out_quantizer", None)
        if out_quantizer or _check_single_layer_block(next_target):
            return [target_layer]

        # Otherwise, we continue searching...
        return _search_block(next_target) + [target_layer]

    def _get_layers_before_dequantizer(model):
        # Return layers before the first dequantizer
        layers = []
        last_deq_layer = None
        for layer in model.layers:
            if isinstance(layer, Dequantizer):
                last_deq_layer = layer
                break
            layers.append(layer)
        return last_deq_layer, layers

    def _is_end_of_block(layer, blocks):
        # Return true if the layer is the end of a new block.
        # The very last layer is the end of another block
        if not blocks:
            return True
        outbounds = layer.outbound_nodes
        # If there are many outbounds, this is an end of block
        if len(outbounds) > 1:
            return True
        # Prepare a list of all first layers of all blocks
        first_layers = [block[0] for block in blocks]
        outbound = outbounds[0].outbound_layer
        return outbound in first_layers

    blocks = []
    # If there is a Dequantizer, save the first one as single block layer and then
    # remove all layers after it (not supported at conversion time)
    deq_layer, model_layers = _get_layers_before_dequantizer(model)
    if deq_layer:
        blocks.append([deq_layer])

    # We forward from head to bottom
    for layer in model_layers[::-1]:
        if _is_end_of_block(layer, blocks):
            new_block = _search_block(layer)
            blocks.append(new_block)
        # The skip layers must be inside of one block
        elif layer not in sum(blocks, []):
            raise RuntimeError(f"Impossible to append {layer.name} into a block.")

    # At the end, we sort blocks from bottom to head
    return blocks[::-1]


def get_block_out_quantizer(block):
    """Extract the output_quantizer of the layers block.

    Note that a block of layers is considered as block if:
        1. There is only one output quantizer (in the last layer of the block)
        2. There are no multi-inbound layers
        3. There are no QuantizedExtractToken layers

    This method should ideally be called after "split_model_into_blocks" to make sure
    that the layers block can be handled correctly.

    Args:
        block (list(:obj:`tf.keras.Layer`)): the layers block.

    Returns:
        :obj:`quantizeml.layers.OutputQuantizer`: the layers block output quantizer (None if the
            block has no output quantizer)
    """
    # Output quantizer is on the last layer of the block
    return getattr(block[-1], "out_quantizer", None)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert QuantizedExtractToken to Akida.
"""
from akida import ExtractToken
import quantizeml.layers as qlayers

from .layer_utils import get_inbound_layers


def _create_extract_token(layer):
    """Parses a quantizeml QuantizedExtractToken layer and returns the
    params to create the corresponding Akida ExtractToken layer.

    Args:
        layer (:obj:`tf.keras.Layer`): the quantizeml QuantizedExtractToken
            layer to convert.

    Returns:
        :obj:`akida.MadNorm`: The created akida layer.
    """
    assert isinstance(layer, qlayers.QuantizedExtractToken)

    # Akida is capable of supporting only a given combination of token
    if isinstance(layer.token, int):
        token_range = [layer.token]
    else:
        token_range = [*layer.token]
    # Check range is continuous
    begin = min(token_range)
    end = max(token_range) + 1
    continuous = sorted(token_range) == list(range(begin, end))
    if not continuous:
        raise ValueError(f"Argument token in {layer.name} should contain a "
                         "continuous range")
    extract_token = ExtractToken(begin=begin,
                                 end=end,
                                 name=layer.name)
    return extract_token


def convert_extract_token(model_ak, layer_k):
    """Converts QuantizedExtractToken layer and adds it to the Akida's model.

    Args:
        model_ak (:obj:`akida.Model`): the Akida model where the model will be
            added.
        layer_k (:obj:`tf.keras.Layer`): the quantizeml QuantizedExtractToken
            layer to convert.
    """
    if not isinstance(layer_k, qlayers.QuantizedExtractToken):
        raise TypeError(f"Layer {layer_k.name} was expected to be "
                        "QuantizedExtractToken")
    # Retrieve the akida inbound layers
    inbound_layers_ak = get_inbound_layers(model_ak, layer_k)
    # Create and add layer to the akida model
    # It is quite simple because there are no variables or params to be set.
    layer_ak = _create_extract_token(layer_k)
    model_ak.add(layer_ak, inbound_layers_ak)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to update akida layer output variables from a keras OutputQuantizer.
"""
import numpy as np
from quantizeml.layers import OutputQuantizer
from quantizeml.tensors import pow2

from ..akida_versions import AkidaVersion, get_akida_version
from .weights import broadcast_and_set_variable


def set_output_variables(layer_ak, out_quantizer):
    """Computes and sets the output variables in an akida layer.

    Args:
        layer_ak (:obj:`akida.Layer`): the targeted akida layer.
        out_quantizer (:obj:`quantizeml.OutputQuantizer`): the source output quantizer.
    """
    assert isinstance(out_quantizer, OutputQuantizer)

    # Extract the OutputQuantizer variables
    scales = out_quantizer.qscales.value.values
    shift = out_quantizer.shift.value

    # Quantizeml evaluates the outputs as: y = x * scales * 2^shift
    if get_akida_version() == AkidaVersion.v1:
        # Calculate the float activation step, as the reciprocal of (scales * 2^shift)
        scales_rec = (pow2(-shift) / scales).numpy().astype(np.float32)
        # In akida 1.0, the outputs are evaluated as: y = x / act_step
        layer_ak.variables["act_step"] = scales_rec
        if layer_ak.parameters.activation:
            # For activations, x is decreased by half the activations step before division
            # to increase the precision. This is obtained by increasing the threshold.
            layer_ak.variables["threshold"] += np.round(0.5 * scales_rec).astype(np.int32)
            # Adjust activation step to match activation formula
            act_bits = layer_ak.parameters.act_bits
            layer_ak.variables["act_step"] *= 2 ** (act_bits - 4)
    else:
        output_scales = scales.numpy().astype(np.uint8)
        broadcast_and_set_variable(layer_ak.variables, "output_scales",
                                   output_scales)
        broadcast_and_set_variable(layer_ak.variables, "output_shift",
                                   shift.numpy().astype(np.int8))



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert QuantizedAttention to Akida.
"""
from akida import LayerType, Attention
import quantizeml.layers as qlayers
import numpy as np

from .weights import broadcast_and_set_variable
from .layer_utils import get_inbound_layers


def _set_attention_variables(ak_layer, k_layer):
    """Computes and sets the variables for an Akida Attention layer.

    This function converts the variables of a Keras layer and sets them into
    the corresponding variables of the equivalent Akida layer.

    Args:
        ak_layer (:obj:`akida.Layer`): the targeted akida layer.
        k_layer (:obj:`tf.keras.Layer`): the source quantized layer.
    """
    assert isinstance(k_layer, qlayers.QuantizedAttention)
    assert ak_layer.parameters.layer_type == LayerType.Attention

    variables_ak = ak_layer.variables

    # Get the QuantizedAttention layer shifts
    value_shift = k_layer.values_shift.value.numpy().astype(np.uint8)
    broadcast_and_set_variable(variables_ak, "value_shift", value_shift)
    shiftmax = k_layer.softmax_op
    shiftmax_input_shift = shiftmax.input_shift.value.numpy()
    if np.any(shiftmax_input_shift < 0):
        raise RuntimeError(
            f"Layer {k_layer.name} contains negative values for " +
            "shiftmax_input_shift, that is not supported")
    broadcast_and_set_variable(ak_layer.variables, "shiftmax_input_shift",
                               shiftmax_input_shift.astype(np.uint8))
    out_quantizer = getattr(k_layer, "out_quantizer", False)
    if out_quantizer:
        output_shift = out_quantizer.shift.value.numpy().astype(np.int8)
        broadcast_and_set_variable(ak_layer.variables, "output_shift",
                                   output_shift)


def _create_attention(layer):
    """Parses a quantizeml QuantizedAttention layer and returns the params to
    create the corresponding Akida Attention layer.

    Args:
        layer (:obj:`tf.keras.Layer`): the quantizeml QuantizedAttention layer to convert.

    Returns:
        :obj:`akida.Attention`: The created akida layer.
    """
    assert isinstance(layer, qlayers.QuantizedAttention)

    # Find out if there is a quantizer
    out_quantizer = getattr(layer, "out_quantizer", False)
    # In quantizeml one bit is reserved automatically for the sign, but in akida
    # this is rather checked during the clipping operations.
    buffer_bits = layer.buffer_bitwidth + 1
    if out_quantizer:
        output_bits = out_quantizer.bitwidth
    else:
        # Default to buffer bitwidth
        output_bits = buffer_bits

    # As with output_bits and buffer_bits, shiftmax_output_bits is set with one
    # bit more.
    shiftmax_output_bits = layer.softmax_op.exp_bitwidth + 1

    return Attention(num_heads=layer.num_heads,
                     output_bits=output_bits,
                     buffer_bits=buffer_bits,
                     shiftmax_output_bits=shiftmax_output_bits,
                     name=layer.name)


def convert_quantized_attention(model_ak, layer_k):
    """Converts QuantizedAttention layer and its variables and adds it to the
    Akida's model.

    Args:
        model_ak (:obj:`akida.Model`): the Akida model where the model will be added.
        layer_k (:obj:`tf.keras.Layer`): the quantizeml QuantizedAttention layer to convert.
    """
    # Retrieve the akida inbound layers
    inbound_layers_ak = get_inbound_layers(model_ak, layer_k)
    if not isinstance(layer_k, qlayers.QuantizedAttention):
        raise TypeError(f"Layer {layer_k.name} was expected to be "
                        "QuantizedAttention")
    layer_ak = _create_attention(layer_k)
    model_ak.add(layer_ak, inbound_layers_ak)
    _set_attention_variables(layer_ak, layer_k)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Utility function referent to padding.
"""
import numpy as np

import quantizeml

from ..model_generator import _get_padding


def get_padding_info(k_layer):
    """From a convolution layer, return the akida.Padding and its padding value.

    Args:
        k_layer (:obj:`tf.keras.Layer`): the source quantized layer.

    Returns:
        tuple: ``akida.Padding`` and the array of padding values.
    """
    assert isinstance(k_layer, (quantizeml.layers.QuantizedConv2D,
                                quantizeml.layers.QuantizedDepthwiseConv2D,
                                quantizeml.layers.QuantizedSeparableConv2D))

    # Padding value must be built in constructor
    padding_ak_value = 0
    if getattr(k_layer, "padding_value", None):
        padding_quantize = k_layer.padding_value_quantizer
        assert isinstance(padding_quantize, quantizeml.layers.AlignedWeightQuantizer)
        padding_value = padding_quantize.qweights.value.values.numpy().astype(np.int32)
        padding_shift = padding_quantize.shift.value.numpy().astype(np.uint8)
        padding_ak_value = padding_value >> padding_shift
    return _get_padding(k_layer.padding), padding_ak_value



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Parsing functions to generate an Akida model from a Keras model quantized with quantizeml api.
"""

import quantizeml.layers as qlayers
from quantizeml.models import record_quantization_variables
from akida import Model

from .depthwise_conv2d import convert_depthwise_conv_block
from .depthwise_conv2d_transpose import convert_depthwise_conv2d_transpose_block
from .dense import convert_dense_block
from .add import convert_quantized_add
from .input_data import convert_input
from .shiftmax import convert_quantized_shiftmax
from .attention import convert_quantized_attention
from .madnorm import convert_quantized_madnorm
from .stem import convert_quantized_stem_layers
from .concatenate import convert_quantized_concatenate
from .convolution import convert_conv_block
from .separable_convolution import convert_sepconv_block
from .batchnorm import convert_batchnorm_block
from .extract_token import convert_extract_token
from .conv2d_transpose import convert_conv_transpose_block
from .dequantizer import convert_dequantizer
from ..akida_versions import AkidaVersion, get_akida_version
from .compatibility_checks import check_model_compatibility, skippable_layers


def generate_model(model, input_is_image):
    """Generates an Akida model.

    This function creates an Akida model by iterating through the layers of the
    quantized model. For each layer, the corresponding akida layer is created and
    added to the Akida model.

    Args:
        model (:obj:`tf.keras.Model`): a Keras model to convert.
        input_is_image (bool): True if input is an 8-bit unsigned tensors (like images).

    Returns:
        :obj:`akida.Model`: the generated Akida model.
    """
    # Check if input model is convertible
    check_model_compatibility(model)

    # First store necessary variables for conversion
    record_quantization_variables(model)

    model_ak = Model()

    # Set index of first layer to be checked
    index = 1
    if isinstance(model.layers[1], qlayers.QuantizedRescaling):
        # ignore rescaling
        index += 1
    # Identify the input block
    ak_version = get_akida_version()
    input_block_size = 0
    if ak_version == AkidaVersion.v2:
        # Convert the akida 2.0 stem block layer if the keras model has one.
        input_block_size = convert_quantized_stem_layers(model_ak, model.layers[index:])
    if input_block_size == 0:
        # Look for an input convolution layer
        input_block_size = convert_conv_block(model_ak, model.layers[index:])
    if input_block_size == 0:
        # Convert the keras InputLayer into an InputData layer
        convert_input(model_ak, model.layers[0], input_is_image)
    index += input_block_size

    # Convert the remaining layers. A while is used to be able to advance if needed.
    layers_size = len(model.layers)

    while index < layers_size:
        layer = model.layers[index]
        # Create and add layer to the akida model
        # The next check converts the potential dense_block layers to a QuantizedDense,
        block_size = convert_dense_block(model_ak, model.layers[index:])
        if block_size > 0:
            index += block_size
            continue
        # conv_block layers to a QuantizedConv2D
        block_size = convert_conv_block(model_ak, model.layers[index:])
        if block_size > 0:
            index += block_size
            continue
        # and sepconv_block to a QuantizedSeparableConv2D
        block_size = convert_sepconv_block(model_ak, model.layers[index:])
        if block_size > 0:
            index += block_size
            continue
        # Dequantizer conversion applies in both version
        if isinstance(layer, qlayers.Dequantizer):
            convert_dequantizer(model_ak, layer)
            index += 1
            continue
        # Skip conversion of this layers set
        if isinstance(layer, skippable_layers):
            index += 1
            continue
        # The other modules are for v2 only
        if get_akida_version() != AkidaVersion.v2:
            # If you got here, the layer is not supported: raise an error.
            raise RuntimeError(f"Layer {layer.name}: unsupported type "
                               f"{layer.__class__.__name__}.")
        block_size = convert_batchnorm_block(model_ak, model.layers[index:])
        if block_size > 0:
            index += block_size
            continue
        block_size = convert_depthwise_conv_block(model_ak, model.layers[index:])
        if block_size > 0:
            index += block_size
            continue
        # convert conv_transpose_block to a QuantizedConv2DTranspose
        block_size = convert_conv_transpose_block(model_ak, model.layers[index:])
        if block_size > 0:
            index += block_size
            continue
        block_size = convert_depthwise_conv2d_transpose_block(model_ak, model.layers[index:])
        if block_size > 0:
            index += block_size
            continue
        if isinstance(layer, qlayers.QuantizedAdd):
            convert_quantized_add(model_ak, layer)
        elif isinstance(layer, qlayers.QuantizedShiftmax):
            convert_quantized_shiftmax(model_ak, layer)
        elif isinstance(layer, qlayers.QuantizedAttention):
            convert_quantized_attention(model_ak, layer)
        elif isinstance(layer, qlayers.QuantizedLayerNormalization):
            convert_quantized_madnorm(model_ak, layer)
        elif isinstance(layer, qlayers.QuantizedConcatenate):
            convert_quantized_concatenate(model_ak, layer)
        elif isinstance(layer, qlayers.QuantizedExtractToken):
            convert_extract_token(model_ak, layer)
        else:
            # If you got here, the layer is not supported: raise an error.
            raise RuntimeError(f"Layer {layer.name}: unsupported type "
                               f"{layer.__class__.__name__}.")
        index += 1

    return model_ak



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert QuantizedDepthwiseConv2D to Akida.
"""
import numpy as np

from quantizeml.layers import (QuantizedDepthwiseConv2D, QuantizedReLU,
                               WeightQuantizer, AlignedWeightQuantizer)
from akida import DepthwiseConv2D

from ..akida_versions import AkidaVersion, get_akida_version
from .activations import set_relu_variables
from .weights import broadcast_and_set_variable
from .outputs import set_output_variables
from .padding import get_padding_info
from .blocks import parse_block_additional_layers


def _set_depthwise_conv2d_variables(layer_ak, layer_k):
    """Computes and sets the variables for an Akida DepthwiseConv2D layer.

    This function converts the variables of a Keras layer and sets them into
    the corresponding variables of the equivalent Akida layer.

    Args:
        layer_ak (:obj:`akida.Layer`): the targeted akida layer.
        layer_k (:obj:`tf.keras.Layer`): the source quantized layer.
    """
    assert isinstance(layer_k, QuantizedDepthwiseConv2D)
    assert isinstance(layer_k.weight_quantizer, WeightQuantizer)
    if layer_k.use_bias:
        assert isinstance(layer_k.bias_quantizer, AlignedWeightQuantizer)

    # Get the weights
    weights = layer_k.weight_quantizer.qweights.value.fp.values.numpy()
    # Flip W and H dimensions for depthwise
    weights = np.flip(weights, axis=[0, 1])

    layer_ak.variables["weights"] = weights.astype(np.int8)

    # Set the bias (if there is one)
    if layer_k.use_bias:
        bias_quantizer = layer_k.bias_quantizer
        bias = bias_quantizer.qweights.value.values.numpy().astype(np.int32)
        bias_shift = bias_quantizer.shift.value.numpy().astype(np.uint8)

        # Unshift the bias and store it
        layer_ak.variables["bias"] = (bias >> bias_shift).astype(np.int8)
        # Also store the bias shift
        broadcast_and_set_variable(layer_ak.variables, "bias_shift", bias_shift)

    # Set input shift
    input_shift = getattr(layer_k, 'input_shift', None)
    if input_shift is not None:
        broadcast_and_set_variable(layer_ak.variables, "input_shift",
                                   layer_k.input_shift.value.numpy().astype(np.uint8))


def _parse_depthwise_conv(layer):
    """Parses a quantizeml.QuantizedDepthwiseConv2D parameters.

    Args:
        layer (:obj:`tf.keras.Layer`): the layer to parse.

    Returns:
        dict: the corresponding akida parameters.
    """
    assert isinstance(layer, QuantizedDepthwiseConv2D)

    # Make sure the DepthwiseConv2D kernel size and stride params are square
    assert layer.kernel_size[0] == layer.kernel_size[1], (
        "DepthwiseConv2D kernel_size should be square")
    assert layer.strides[0] == layer.strides[1], (
        "DepthwiseConv2D strides should be the same on both dimension")

    # Padding value must be built in constructor
    padding, _ = get_padding_info(layer)

    # The only weight bitwidth supported is [4, 8]
    weight_bits = layer.weight_quantizer.bitwidth
    assert weight_bits in [4, 8]

    # In quantizeml one bit is reserved for the sign in the buffer bitwidth
    # variable, but in akida this value has to be added back to have the
    # correct clipping.
    buffer_bits = layer.buffer_bitwidth + 1

    return dict(
        kernel_size=layer.kernel_size[0],
        kernel_stride=layer.strides[0],
        padding=padding,
        buffer_bits=buffer_bits,
        activation=False,
        name=layer.name
    )


def convert_depthwise_conv_block(model_ak, layers):
    """Converts a depthwise convolutional block into an akida DepthwiseConv2D layer.

    The expected sequence is:

    - QuantizedDepthwiseConv2D,
    - QuantizedMaxPool2D (optional),
    - QuantizedReLU (optional).

    Args:
        model_ak (:obj:`akida.Model`): the target Akida model.
        layers (list(:obj:`tf.keras.Layer`)): the block layers.

    Return:
        int: the number of layers in the block or O if the first layer is not a
        QuantizedDepthwiseConv2D.
    """
    if len(layers) == 0 or not isinstance(layers[0], QuantizedDepthwiseConv2D):
        return 0

    if get_akida_version() == AkidaVersion.v1:
        return 0

    dw_conv = layers[0]

    # Evaluate the depthwise convolutional layer parameters
    conv_params = _parse_depthwise_conv(dw_conv)

    # Identify the next layers
    next_layers = parse_block_additional_layers(layers, conv_params)
    # Check if we have ReLU or MaxPool
    relu_layer = next_layers.get(QuantizedReLU, None)

    if relu_layer:
        # Output quantizer should be in ReLU layer
        out_quantizer = getattr(relu_layer, "out_quantizer", False)
    else:
        # otherwise, output quantize is in the conv layer
        out_quantizer = getattr(dw_conv, "out_quantizer", False)

    # add output quantizer bitwidth parameter
    if out_quantizer:
        conv_params["output_bits"] = out_quantizer.bitwidth
    else:
        conv_params["output_bits"] = conv_params["buffer_bits"]

    # Create Akida layer
    dw_conv_ak = DepthwiseConv2D(**conv_params)
    # Add layer to the model to build its internal variables
    model_ak.add(dw_conv_ak)

    # Set base variables
    _set_depthwise_conv2d_variables(dw_conv_ak, dw_conv)

    # Set optional activation variables
    if relu_layer:
        set_relu_variables(dw_conv_ak, relu_layer)

    # Set optional output_quantizer variables
    if out_quantizer:
        set_output_variables(dw_conv_ak, out_quantizer)

    return 1 + len(next_layers)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert keras pooling layers parameters to akida.
"""

from quantizeml.layers import QuantizedMaxPool2D, QuantizedGlobalAveragePooling2D, OutputQuantizer
from akida import Padding, PoolType
from ..akida_versions import AkidaVersion, get_akida_version


def parse_max_pooling(layer):
    """Parses a quantizeml.QuantizedMaxPool2D parameters.

    Args:
        layer (:obj:`quantizeml.QuantizedMaxPool2D`): the layer to parse.

    Returns:
        dict: the corresponding akida parameters.
    """
    assert isinstance(layer, QuantizedMaxPool2D)

    padding = Padding.Same if layer.padding == "same" else Padding.Valid
    pool_size = layer.pool_size
    pool_stride = layer.strides if layer.strides else pool_size

    pool_params = dict(
        pool_type=PoolType.Max,
        padding=padding,
        pool_size=pool_size,
        pool_stride=pool_stride
    )

    if get_akida_version() == AkidaVersion.v2:

        # Make sure the Conv2D layers spatial params are square
        assert pool_size[0] == pool_size[1], "Akida v2 layers only handle square pooling"
        pool_params.update({"pool_size": pool_size[0]})
        assert pool_stride[0] == pool_stride[1], (
            "Akida v2 layers pooling stride should be the same for both dimensions")
        pool_params.update({"pool_stride": pool_stride[0]})

    return pool_params


def parse_global_average_pooling(layer):
    """Parses a quantizeml.QuantizedGlobalAveragePooling2D parameters.

    Args:
        layer (:obj:`quantizeml.QuantizedGlobalAveragePooling2D`): the layer to parse.

    Returns:
        dict: the corresponding akida parameters.
    """
    assert isinstance(layer, QuantizedGlobalAveragePooling2D)

    if get_akida_version() == AkidaVersion.v1:
        return dict(pool_type=PoolType.Average)

    # Check if there is an output_quantizer
    out_quantizer = getattr(layer, "out_quantizer", None)

    assert isinstance(out_quantizer, OutputQuantizer)

    return dict(pool_type=PoolType.Average, output_bits=out_quantizer.bitwidth)


def set_pooling_variables(layer_ak, layer_k):
    """Computes and sets the pooling variables in an akida layer.

    Args:
        layer_ak (:obj:`akida.Layer`): the targeted akida layer.
        layer_k (:obj:`quantizeml.Layer`): the source layer.
    """
    assert isinstance(layer_k, (QuantizedGlobalAveragePooling2D, QuantizedMaxPool2D))
    # Nothing to do in AkidaVersion.v2 or for QuantizedMaxPool2D
    if get_akida_version() == AkidaVersion.v2 or isinstance(layer_k, QuantizedMaxPool2D):
        return

    # In AkidaVersion.v1, biases are only applied when evaluating activations, by
    # subtracting the threshold. Since this happens after pooling, the scaling factor
    # introduced by the GAP operation must be applied to the threshold.
    spatial_size = layer_k.input_shape[1] * layer_k.input_shape[2]
    if spatial_size > 2**8:
        raise RuntimeError(f"Unsupported spatial size product ({spatial_size}). "
                           "We only accepts values smaller than 256. "
                           "Consider to replace GlobalAveragePooling2D by MaxPooling2D.")
    layer_ak.variables["threshold"] *= spatial_size



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert QuantizedConv2D to Akida.
"""
import numpy as np

from quantizeml.layers import (QuantizedConv2D, QuantizedReLU, QuantizedGlobalAveragePooling2D,
                               WeightQuantizer, AlignedWeightQuantizer)
from akida import InputConvolutional, Convolutional, LayerType, Conv2D, InputConv2D

from ..akida_versions import AkidaVersion, get_akida_version
from .activations import set_relu_variables
from .pooling import set_pooling_variables
from .outputs import set_output_variables
from .padding import get_padding_info
from .blocks import parse_block_additional_layers, get_block_out_quantizer
from .weights import broadcast_and_set_variable


def _set_conv2d_variables(layer_ak, layer_k):
    """Computes and sets the variables for an Akida convolutional layer.

    This function converts the variables of a Keras layer and sets them into
    the corresponding variables of the equivalent Akida layer.

    Args:
        layer_ak (:obj:`akida.Layer`): the targeted akida layer.
        layer_k (:obj:`tf.keras.Layer`): the source quantized layer.
    """
    assert isinstance(layer_k, QuantizedConv2D)
    assert isinstance(layer_k.weight_quantizer, WeightQuantizer)
    if layer_k.use_bias:
        assert isinstance(layer_k.bias_quantizer, AlignedWeightQuantizer)

    # Get the weights
    weights = layer_k.weight_quantizer.qweights.value.fp.values.numpy()
    if layer_ak.parameters.layer_type in (LayerType.Convolutional, LayerType.Conv2D):
        # Flip W and H dimensions for conv. kernels (not input conv.)
        weights = np.flip(weights, axis=[0, 1])
    layer_ak.variables["weights"] = weights.astype(np.int8)

    # Set the bias (if there is one)
    if layer_k.use_bias:
        bias_quantizer = layer_k.bias_quantizer
        bias = bias_quantizer.qweights.value.values.numpy().astype(np.int32)
        bias_shift = bias_quantizer.shift.value.numpy().astype(np.uint8)
        target_version = get_akida_version()
        if target_version == AkidaVersion.v1:
            # Store bias into the threshold variable
            layer_ak.variables["threshold"] = -bias
        else:
            # Unshift the bias and store it
            layer_ak.variables["bias"] = (bias >> bias_shift).astype(np.int8)
            # Also store the bias shift
            broadcast_and_set_variable(layer_ak.variables, "bias_shift", bias_shift)


def _parse_convolutional(layer):
    """Parses a quantizeml.QuantizedConv2D parameters.

    Args:
        layer (:obj:`tf.keras.Layer`): the layer to parse.

    Returns:
        dict: the corresponding akida parameters.
    """
    assert isinstance(layer, QuantizedConv2D)

    # Padding value must be built in constructor
    padding, padding_value = get_padding_info(layer)

    # The only weight bitwidth supported is [4, 8]
    weight_bits = layer.weight_quantizer.bitwidth
    assert weight_bits in [4, 8]

    conv_params = dict(
        input_shape=tuple(int(x) for x in layer.input_shape[1:4]),
        padding=padding,
        padding_value=padding_value,
        kernel_size=(layer.kernel_size[0], layer.kernel_size[1]),
        filters=int(layer.kernel.shape[3]),
        weights_bits=layer.weight_quantizer.bitwidth,
        kernel_stride=(layer.strides[0], layer.strides[1]),
        activation=False,
        name=layer.name
    )

    if get_akida_version() == AkidaVersion.v2:
        # Conv2D layer has no weight bits, nor padding_value params
        conv_params.pop("weights_bits", None)

        # Make sure the Conv2D layers spatial params are square
        assert layer.kernel_size[0] == layer.kernel_size[1], ("Conv2D and InputConv2D handle only"
                                                              "square kernels")
        conv_params.update({"kernel_size": layer.kernel_size[0]})
        assert layer.strides[0] == layer.strides[1], ("For Conv2D and InputConv2D stride should be"
                                                      "the same for both dimensions")
        conv_params.update({"kernel_stride": layer.strides[0]})

        # In quantizeml one bit is reserved for the sign in the buffer bitwidth
        # variable, but in akida this value has to be added back to have the
        # correct clipping.
        buffer_bits = layer.buffer_bitwidth + 1
        conv_params["buffer_bits"] = buffer_bits
        # Find out if there is a quantizer
        out_quantizer = getattr(layer, "out_quantizer", False)

        if out_quantizer:
            conv_params["output_bits"] = out_quantizer.bitwidth
        else:
            conv_params["output_bits"] = buffer_bits
    return conv_params


def convert_conv_block(model_ak, layers):
    """Converts an convolutional block into an akida convolutional layer.

    The expected sequence is:

    - QuantizedConv2D,
    - QuantizedMaxPooling2D (optional),
    - QuantizedRelU (optional).

    Args:
        model_ak (:obj:`akida.Model`): the target Akida model.
        layers (list(:obj:`tf.keras.Layer`)): the block layers.

    Return:
        int: the number of layers in the block or O if the first layer is not a QuantizedConv2D.
    """
    if len(layers) == 0 or not isinstance(layers[0], QuantizedConv2D):
        return 0

    conv = layers[0]

    # Evaluate the convolutional layer parameters
    conv_params = _parse_convolutional(conv)
    next_layers = parse_block_additional_layers(layers, conv_params)

    if len(model_ak.layers) == 0 and conv_params["input_shape"][-1] in [1, 3]:
        if get_akida_version() == AkidaVersion.v1:
            conv_ak = InputConvolutional(**conv_params)
        else:
            conv_ak = InputConv2D(**conv_params)
    elif len(model_ak.layers) != 0:
        # A Convolutional should be added if there is a previous layer (like InputData).
        # Unlike InputConvolutional or InputConv2D, Convolutional and Conv2D do not receive
        #  the following parameters:
        for no_param in ["input_shape", "padding_value"]:
            conv_params.pop(no_param, None)
        if get_akida_version() == AkidaVersion.v1:
            conv_ak = Convolutional(**conv_params)
        else:
            conv_ak = Conv2D(**conv_params)
    else:
        # Convert loop will manage the other cases
        return 0

    # Add layer to the model to build its internal variables
    model_ak.add(conv_ak)
    # Set base variables
    _set_conv2d_variables(conv_ak, conv)

    # Check if we have ReLU or GAP
    pool_layer = next_layers.get(QuantizedGlobalAveragePooling2D, None)
    if pool_layer:
        set_pooling_variables(conv_ak, pool_layer)
    relu_layer = next_layers.get(QuantizedReLU, None)
    if relu_layer:
        if get_akida_version() == AkidaVersion.v2:
            # Set the optional activation variables
            set_relu_variables(conv_ak, relu_layer)

    # Get out_quantizer of the block.
    out_quantizer = get_block_out_quantizer([conv] + list(next_layers.values()))

    if out_quantizer:
        set_output_variables(conv_ak, out_quantizer)

    return 1 + len(next_layers)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert QuantizedSeparableConv2D to Akida.
"""
import numpy as np

from akida import LayerType, SeparableConvolutional
from quantizeml.layers import (QuantizedSeparableConv2D, QuantizedGlobalAveragePooling2D,
                               WeightQuantizer, AlignedWeightQuantizer)

from ..akida_versions import AkidaVersion, get_akida_version
from .pooling import set_pooling_variables
from .outputs import set_output_variables
from .padding import get_padding_info
from .blocks import parse_block_additional_layers, get_block_out_quantizer


def _set_sepconv_variables(ak_layer, k_layer):
    """Computes and sets the variables for an Akida SeparableConvolutional layer.

    This function converts the variables of a Keras layer and sets them into
    the corresponding variables of the equivalent Akida layer.

    Args:
        ak_layer (:obj:`akida.Layer`): the targeted akida layer.
        k_layer (:obj:`tf.keras.Layer`): the source quantized layer.
    """
    assert isinstance(k_layer, QuantizedSeparableConv2D)
    assert ak_layer.parameters.layer_type == LayerType.SeparableConvolutional
    assert isinstance(k_layer.dw_weight_quantizer, WeightQuantizer)
    assert isinstance(k_layer.pw_weight_quantizer, WeightQuantizer)

    variables_ak = ak_layer.variables

    # Get the QuantizedSeparableConv2D weights
    weights_ak = k_layer.dw_weight_quantizer.qweights.value.fp.values.numpy()
    weights_pw_ak = k_layer.pw_weight_quantizer.qweights.value.fp.values.numpy()
    # We require flip depthwise weights
    weights_ak = np.flip(weights_ak, axis=[0, 1])

    # Get the QuantizedSeparableConv2D bias
    if k_layer.use_bias:
        bias_quantizer = k_layer.bias_quantizer
        assert isinstance(bias_quantizer, AlignedWeightQuantizer)
        bias = bias_quantizer.qweights.value.values.numpy().astype(np.int32)
        # Store bias into the threshold variable
        variables_ak["threshold"] = -bias

    variables_ak["weights"] = weights_ak.astype(np.int8)
    variables_ak["weights_pw"] = weights_pw_ak.astype(np.int8)


def _parse_sepconv(layer):
    """Parses a quantizeml.SeparableConvolutional parameters.

    Args:
        layer (:obj:`tf.keras.Layer`): the layer to parse.

    Returns:
        dict: the corresponding akida parameters.
    """

    assert isinstance(layer, QuantizedSeparableConv2D)

    # The only weight bitwidth supported is 4
    weight_bits = layer.dw_weight_quantizer.bitwidth
    assert weight_bits == 4 and weight_bits == layer.pw_weight_quantizer.bitwidth

    # Padding value must be built in constructor
    padding, _ = get_padding_info(layer)

    sepconv_params = dict(kernel_size=(layer.kernel_size[0], layer.kernel_size[1]),
                          filters=int(layer.pointwise_kernel.shape[3]),
                          padding=padding,
                          kernel_stride=(layer.strides[0], layer.strides[1]),
                          weights_bits=weight_bits,
                          activation=False,
                          name=layer.name)
    return sepconv_params


def convert_sepconv_block(model_ak, layers):
    """Converts a separable convolutional block into an akida SeparableConvolutional layer.

    The expected sequence is:

        - QuantizedSeparableConv2D,
        - QuantizedMaxPooling2D (optional),
        - QuantizedReLU (optional).

    Args:
        model_ak (:obj:`akida.Model`): the Akida model where the model will be added.
        layers (list(:obj:`tf.keras.Layer`)): the remaining model layers to convert.

    Return:
        int: the number of layers in the block or O if the first layer is not a
            QuantizedSeparableConv2D.
    """
    if len(layers) == 0 or not isinstance(layers[0], QuantizedSeparableConv2D):
        return 0

    if get_akida_version() == AkidaVersion.v2:
        return 0

    if len(model_ak.layers) == 0:
        raise RuntimeError("An InputData layer is required previous to add a new SeparableConv2D.")

    sepconv = layers[0]

    # Evaluate the separable convolutional layer parameters
    sepconv_params = _parse_sepconv(sepconv)
    next_layers = parse_block_additional_layers(layers, sepconv_params)

    # Create the Akida SeparableConv2D layer
    sepconv_ak = SeparableConvolutional(**sepconv_params)

    # Add layer to the model to build its internal variables
    model_ak.add(sepconv_ak)

    # Set variables
    _set_sepconv_variables(sepconv_ak, sepconv)
    pool_layer = next_layers.get(QuantizedGlobalAveragePooling2D, None)
    if pool_layer:
        set_pooling_variables(sepconv_ak, pool_layer)
    out_quantizer = get_block_out_quantizer([sepconv] + list(next_layers.values()))
    if out_quantizer:
        set_output_variables(sepconv_ak, out_quantizer)

    return 1 + len(next_layers)
